﻿/*
-				Map Editor Script
-				Created By:
-				Doom_Killer
-
-				Version:
-				1.2.5
-
-				Credits:
-				SLC - For providing a better SQLite plugin.
-				Juppi - Free Camera Movement Script.
-
-				Enjoy your mapping, if you find any bugs, report them on VCMP forum on the script's topic.
-*/
/*-				Changelogs	
-*/
/*
-				Changelog - v1.1
-				► Changed data-type for rotations x, y, z to FLOAT.
-
-				Changelog - v1.2	
-				► Improved queries to reduce memory leakage.
-				► Added Page UP | DOWN key in keys list.
-				
-				Changelog - v1.2.1
-				► Fixed error on location returning.
-				
-				Changelog - v1.2.2
-				► Fixed error on newmap, unable to release sqlite data.
-				► Fixed error when not sending any text argument in addobject.
-
-				Changelog - v1.2.3
-				► Fixed errors.
-				► Fixed crash on loading map.
-				► Fixed some other things regarding errors and crashes.
-
-				Changelog - v1.2.4
-				► Now with free camera view! /cam to toggle!
-
-				Changelog - v1.2.5
-				► Added text for Currently Holding Object Name and Model.
-				► Added Auto Saving for maps, default auto save every minute.
-				► Added autosave command! /autosave to toggle.
-				
-				Changelog - v1.2.6
-				► Map menu is now available! use /addobject without any id to enter map menu and select map while previewing!
-				► Press CTRL to select & create the object.
-				
*/

// =========================================== CONSTANTS ==============================================

const Motion_Speed = 100; // You should not touch this unless you know what you're doing.
const cAutoSaveTime = 60000; // Auto save every minute, you can change the time here.

// =========================================== CLASES =================================================

class PlayerData
{
	MapDelete = false;
	PositionMode = false;
	Speed = "normal";
	Editing = false;
	Holding = false;
	Motion = null;
	LastKey = null;
	CTRL = false;
	car = null;
	txdObject = null;
	// Object selection.
	lastPos = null;
	viewObject = null;
	viewModel = 500;
	motionTimer = null;
}

// =========================================== LOCALS ==================================================

local txdCurrentMap;
local txdTotalObjects;
local autoSave = true;

// =========================================== SERVER EVENTS ===========================================

/*
function onServerStart()
{
}

function onServerStop()
{
}
*/

function onScriptLoad()
{
	// =========================================== ARRAYS ==============================================
	pData <- array( GetMaxPlayers(), null );
	Object <- array( GetMaxPlayers(), null );
	
	// =========================================== DATABASE ============================================
	Maps <- SQLite_Open( "scripts/Maps.sqlite" );
	db <- SQLite_Open( "scripts/locations.sqlite" );
	
	// =========================================== GLOBAL ==============================================
	CurrentMap <- "";
	objCreated <- 0;
	
	// =========================================== FILES ===============================================
	dofile( "scripts/Freeview.addon" );
	onCamScriptLoad();
	
	txdCurrentMap = CreateTextdraw( "Current Map: None", 1, 735, 0xFFFFFF00 );
	txdCurrentMap.SetRelativeForAll( true );
	
	txdTotalObjects = CreateTextdraw( "Total Objects Created: " + objCreated, 1, 720, 0xFF00FFFF );
	txdTotalObjects.SetRelativeForAll( true );
	
	// =========================================== BINDS ===============================================
	enter <- BindKey( true, 0x0D, 0, 0 );
	del <- BindKey( true, 0x2E, 0, 0 );
	backspace <- BindKey( true, 0x08, 0, 0 );
	R <- BindKey( true, 0x52, 0, 0 );
	one <- BindKey( true, 0x31, 0, 0 );
	two <- BindKey( true, 0x32, 0, 0 );
	ctrl <- BindKey( true, 0x11, 0, 0 );
	c <- BindKey(true, 0x43, 0, 0 );
	PageUp <- BindKey( true, 0x21, 0, 0 );
	PageDown <- BindKey( true, 0x22, 0, 0 );
	ArrowUp <- BindKey( true, 0x26, 0, 0 );
	ArrowDown <- BindKey( true, 0x28, 0, 0 );
	ArrowLeft <- BindKey( true, 0x25, 0, 0 );
	ArrowRight <- BindKey( true, 0x27, 0, 0 );
	
	// =========================================== TIMERS ==============================================
	NewTimer( "AutoSave", cAutoSaveTime, 0 );
	
	// =========================================== PRINTS ==============================================
	print( "Doom_Killer's Map Editor successfully started!" );
	print( "Editor successfully running! ");
}

function onScriptUnload()
{

}

// =========================================== PLAYER EVENTS ===========================================

function onPlayerJoin( player )
{
	onCamPlayerJoin( player )
	pData[ player.ID ] = PlayerData( player.Name );
	Object[ player.ID ] = null;
	txdCurrentMap.ShowForPlayer( player );
	txdTotalObjects.ShowForPlayer( player );
	MessagePlayer( "[#ffffff][Server]: [#00CC00]Welcome to Map Editor!", player );
}

function onPlayerPart( player, reason )
{
	onCamPlayerPart( player, reason )
	pData[ player.ID ].Editing = false;
	pData[ player.ID ] = null;
	Object[ player.ID ] = null;
}

function onPlayerSpawn( player )
{
	NewTimer( "ClientRender", 1000, 0, player.ID );
	
	player.Pos = Vector( -335.591, -524.865, 12.7615 );
	player.Immunity = 31;
	player.GiveWeapon( 32, 99999 );
	
	MessagePlayer( "[#00CC00][KEYS]: [#FFFFFF]These are the keys used to interact with the editor!", player );
	MessagePlayer( "[#00CC00][KEYS]: [#FFFFFF]Arrow LEFT | RIGHT | UP | DOWN - Movement & Rotations", player );
	MessagePlayer( "[#00CC00][KEYS]: [#FFFFFF]Page UP | DOWN - Z Movement/Rotation.", player );
	MessagePlayer( "[#00CC00][KEYS]: [#FFFFFF]CTRL + C - Clone Object | R - Reset Object rotations | Delete - Delete object", player );
	MessagePlayer( "[#00CC00][KEYS]: [#FFFFFF]1 - Switch between Rotation & Position Mode | 2 - Speed variation", player );
	MessagePlayer( "[#00CC00][KEYS]: [#FFFFFF]Backspace - Stop controlling an object", player );
	MessagePlayer( "[#00CC00][COMMAND]: [#FFFFFF]/cam - To enable/disable free-view.", player );
}

function onPlayerCommand( player, cmd, text )
{
	if( cmd != "cam" && pCamera[ player.ID ].IsEnabled() ) return MessagePlayer( "[#00CC00][ERROR]: [#FFFFFF]Other commands are disabled when in camera mode!", player );
	
	if ( cmd == "cam" )
	{
		if( pData[ player.ID ].lastPos != null )
		{
			return null;
		}
		else if ( !pCamera[ player.ID ].IsEnabled() )
		{
			pCamera[ player.ID ].Enable();
			pData[ player.ID ].Editing = false;
			Object[ player.ID ] = null;
			MessagePlayer( "[#00CC00][CAM]: [#ffffff]Free view camera [#00CC00]enabled[#ffffff]. Type '/cam' to disable.", player );
		}
		else
		{
			pCamera[ player.ID ].Disable();
			MessagePlayer( "[#00CC00][CAM]: [#ffffff]Free view camera [#00CC00]disabled[#ffffff].", player );
		}
	}
	
	else if( cmd == "s" )
	{
		if( !text ) { print( "Pos: " + player.Pos ); return MessagePlayer( "Position saved!", player ); }
		print( "Pos: " + player.Pos + " //" + text );
		MessagePlayer( "Position " + text + "saved!", player );
	}
	
	else if( cmd == "newmap" )
	{
		if( text && CurrentMap == "" )
		{
			if( text == "default" ) return MessagePlayer( "[#00CC00][ERROR]: [#FFFFFF]The name [#00CC00]default [#FFFFFF]cannot be used to create a map", player );
			local check_map;
			try { check_map = SQLite_Query( Maps, "SELECT * FROM sqlite_master WHERE type='table' AND name='" + CurrentMap.tolower() + "'"); } catch( e ) return MessagePlayer( "[#00CC00][MAP]: [#ffffff]A map with that name already exists!", player );
			SQLite_Exec( Maps, "CREATE TABLE '" + text.tolower() + "'( model INT, x INT, y INT, z INT, rotx FLOAT, roty FLOAT, rotz FLOAT )" );
			CurrentMap = "" + text.tolower();
			Message( "[#ffffff][New-Map]: [#00CC00]A new map [#ffffff]" + CurrentMap + " [#00CC00]was started by [#ffffff]" + player.Name + "." );
			if( check_map != null ) SQLite_Release( check_map );
		}
		else if( text && CurrentMap != "" )
		{ 
			Message( "[#00CC00][NEWMAP]: [#ffffff]The map [#00CC00]" + CurrentMap + " [#ffffff]is still loaded, please close this map before proceeding. /savemap to save & close!" );
		}
	}
	
	else if( cmd == "addobject" )
	{
		if( CurrentMap == "" ) return MessagePlayer( "[#00CC00][MAP]: [#ffffff]You must open a map to add objects!", player );
		else if( !text )
		{
			player.IsFrozen = true;
			player.World = 100 + player.ID;
			pData[ player.ID ].viewObject = CreateObject( pData[ player.ID ].viewModel, player.World, -1372.88, 1282.51, 83.8704, 255 );
			pData[ player.ID ].motionTimer = NewTimer( "onObjectRender", 100, 0, player.ID );
			pData[ player.ID ].lastPos = player.Pos;
			player.Pos = Vector( -1351.37, 1275.35, 91.1706 );
			player.SetCameraPos( Vector( -1351.37, 1275.35, 91.1706 ), Vector( -1372.88, 1282.51, 83.8704 ) );
		}
		else if( text && text != "" && IsNum( text ) )
		{
			Object[ player.ID ] = CreateObject( text.tointeger(), 0, player.Pos.x, player.Pos.y, player.Pos.z, 255 );
			Object[ player.ID ].TrackingShots = true;
			objCreated++;
			player.Pos = Vector( player.Pos.x, player.Pos.y, player.Pos.z + 3 );
			pData[ player.ID ].Editing = true;
			Message( "[#00CC00][MAP]: [#ffffff]Created object " + text );
		}
	}
	
	else if( cmd == "map" )
	{
		if( CurrentMap != "" ) Message( "[#00CC00][MAP]: [#ffffff]" + CurrentMap );
		else Message( "[#00CC00][MAP]: [#ffffff]None" );
	}
	
	else if( cmd == "savemap" )
	{
		if( CurrentMap != "" )
		{
			if( objCreated == 0 )
			{
				SQLite_Exec( Maps, "DROP TABLE " + CurrentMap.tolower() + "" );
				CurrentMap = "";
				return MessagePlayer( "[#ffffff][Map]: [#00CC00]The map has been [#ffffff]closed & deleted [#00CC00]as no objects were created.", player );
			}
			
			SQLite_Exec( Maps, "DELETE FROM '" + CurrentMap.tolower() + "'" );
			pData[ player.ID ].Editing = false;
			for( local i = 0; i < objCreated; i++ )
			{
				local obj = FindObject(i);
				if( obj )
				{
					SQLite_Exec( Maps, "INSERT INTO '" + CurrentMap.tolower() + "' VALUES ( '" + obj.Model + "', '" + obj.Pos.x + "', '" + obj.Pos.y + "', '" + obj.Pos.z + "', '" + obj.RotationEuler.x + "', '" + obj.RotationEuler.y + "', '" + obj.RotationEuler.z + "');" );
				}
			}
			Message( "[#00CC00][MAP]: [#ffffff]" + player.Name + " has saved the map" );
		}
	}
	
	else if( cmd == "closemap" )
	{
		if( CurrentMap != "" )
		{
			if( objCreated == 0 )
			{
				SQLite_Exec( Maps, "DROP TABLE " + CurrentMap.tolower() );
				CurrentMap = "";
				return MessagePlayer( "[#ffffff][Map]: [#00CC00]The map has been [#ffffff]closed & deleted [#00CC00]as no objects were created.", player );
			}
			
			pData[ player.ID ].Editing = false;
			Object[ player.ID ] = null;
			for( local i = 0; i < objCreated; i++ )
			{
				local obj = FindObject(i);
				if( obj ) obj.Delete();
			}
			CurrentMap = "";
			objCreated = 0;
			Message( "[#00CC00][MAP]: [#ffffff]" + player.Name + " has closed the map" );
		}
	}
	
	else if( cmd == "exportmap" )
	{
		if( CurrentMap != "" )
		{
			if( objCreated == 0 )
			{
				SQLite_Exec( Maps, "DROP TABLE " + CurrentMap.tolower() + "" );
				CurrentMap = "";
				return MessagePlayer( "[#ffffff][Map]: [#00CC00]The map has been [#ffffff]closed & deleted [#00CC00]as no objects were created.", player );
			}
			
			SQLite_Exec( Maps, "DELETE FROM '" + CurrentMap.tolower() + "'" );
			pData[ player.ID ].Editing = false;
			for( local i = 0; i < objCreated; i++ )
			{
				local obj = FindObject(i);
				if( obj )
				{
					SQLite_Exec( Maps, "INSERT INTO '" + CurrentMap.tolower() + "' VALUES ( '" + obj.Model + "', '" + obj.Pos.x + "', '" + obj.Pos.y + "', '" + obj.Pos.z + "', '" + obj.RotationEuler.x + "', '" + obj.RotationEuler.y + "', '" + obj.RotationEuler.z + "');" );
					print( "CreateObject( " + obj.Model + ", 1, Vector( " + obj.Pos.x + ", " + obj.Pos.y + ", " + obj.Pos.z + "), 255 ).RotateByEuler( Vector( " + obj.RotationEuler.x + ", " + obj.RotationEuler.y + ", " + obj.RotationEuler.z + "), 0);" );
					obj.Delete();
				}
			}
			CurrentMap = "";
			objCreated = 0;
			Object[ player.ID ] = null;
			Message( "[#00CC00][MAP]: [#ffffff]" + player.Name + " has exported the map!" );
		}
	}
	
	else if( cmd == "loadmap" )
	{
		if( CurrentMap != "" ) MessagePlayer( "[#00CC00][MAP]: [#ffffff]A map is already opened!", player );
		else
		{
			if( !text ) return MessagePlayer( "[#00CC00][MAP]: [#ffffff]Enter map name to load!", player );
			else
			{
				local map;
				try
				{
					map = SQLite_Query( Maps, "SELECT * FROM '" + text.tolower() + "'" );
				}
				catch( e ) return MessagePlayer( "[#00CC00][MAP]: [#ffffff]Map not found!", player );
				
				do
				{
					local model = SQLite_ColumnData( map, 0 );
					local x = SQLite_ColumnData( map, 1 );
					local y = SQLite_ColumnData( map, 2 );
					local z = SQLite_ColumnData( map, 3 );
					local rotx = SQLite_ColumnData( map, 4 );
					local roty = SQLite_ColumnData( map, 5 );
					local rotz = SQLite_ColumnData( map, 6 );
					local _obj = CreateObject( model, 1, x, y, z, 255 );
					_obj.RotationEuler.x = rotx.tofloat();
					_obj.RotationEuler.y = roty.tofloat();
					_obj.RotationEuler.z = rotz.tofloat();
					objCreated++;
					_obj.TrackingShots = true;
					Object[ player.ID ] = null;
					//Object[ player.ID ] = _obj;
					//Message( "[#00CC00][LOAD]: [#ffffff]Created Object at " + x + ", " + y + ", z with rotation " + rotx + ", " + roty + ", " + rotz );
				}
				while ( SQLite_NextRow( map ) );
				SQLite_Release( map );
			}
			CurrentMap = "" + text.tolower(); 
			Message( "[#00CC00][MAP]: [#ffffff]" + text + " [#00CC00]has been loaded by [#ffffff]" + player.Name + "." );
		}
	}	
	
	else if( cmd == "desall" )
	{
		if( CurrentMap != "" ) RemoveObjects( CurrentMap );
	}
	
	else if( cmd == "delobj" )
	{
		if( !pData[ player.ID ].MapDelete )
		{
			pData[ player.ID ].MapDelete = true;
			MessagePlayer( "[#00CC00][DELETE]: [#ffffff]Map delete mode is [#00ff00]ON[#ffffff]! Touch object to delete it!", player );
			EnableBumpTracks();
		}
		
		else if( pData[ player.ID ].MapDelete )
		{
			pData[ player.ID ].MapDelete = false;
			MessagePlayer( "[#00CC00][DELETE]: [#ffffff]Map delete mode is [#ff0000]OFF[#ffffff]!", player );
			DisableBumpTracks();
		}
	}
	
	else if( cmd == "car" )
	{
		if( text && IsNum( text ) )
		{
			local veh = text.tointeger();
			if( veh >= 130 && veh <= 236 )
			{
				local pPos = player.Pos;
				if( pData[ player.ID ].car != null ) pData[ player.ID ].car.Remove();
				pData[ player.ID ].car = CreateVehicle( veh, player.World, player.Pos, player.Angle, -1, -1 );
				player.Vehicle = pData[ player.ID ].car;
				ClientMessage( "[#00CC00][CAR]: [#ffffff]Your vehicle has been spawned", player, 255,255,255 );
			}
			else ClientMessage( "[#00CC00][ERROR]: [#ffffff]Invalid vehicle ID", player, 255,255,255 );
		}
				
		else if( text && !IsNum( text ) )
		{
			local veh = GetVehicleModelFromName( text );
			if( veh >= 130 && veh <= 236 )
			{
				local pPos = player.Pos;
				if( pData[ player.ID ].car != null ) pData[ player.ID ].car.Remove();
				pData[ player.ID ].car = CreateVehicle( veh, player.World, player.Pos, player.Angle, -1, -1 );
				player.Vehicle = pData[ player.ID ].car;
				ClientMessage( "[#00CC00][CAR]: [#ffffff]Your vehicle has been spawned", player, 255,255,255 );
			}
		}
	}
	
	else if( cmd == "exec" ) 
	{
		if( !text ) ClientMessage( "[#00CC00][ERROR]: [#ffffff]/exec [Code]", player, 255,255,255 );
		else
		{
			try
			{
				local script = compilestring( text );
				script();
				MessagePlayer( "Done", player );
			}
			catch(e) MessagePlayer( "Error: " + e, player);
		}
	}
	
	else if( cmd == "autosave" )
	{
		switch( autoSave )
		{
			case true:
			autoSave = false;
			Message( "[#00CC00][AUTO-SAVE]: [#FFFFFF]Auto-Save has been [#ff0000]disabled[#FFFFFF] by " + player.Name );
			break;
			
			case false:
			autoSave = true;
			Message( "[#00CC00][AUTO-SAVE]: [#FFFFFF]Auto-Save has been [#00ff00]enabled[#FFFFFF] by " + player.Name );
			break;
		}
	}
	
	else if ( cmd == "saveloc" )
	{
		if ( !text ) MessagePlayer( "Error - Usage: / saveloc <location name>", player );
		else if ( !player.IsSpawned ) MessagePlayer( "Error - You haven't spawned.", player );
        else if ( FindLoc( text ) ) MessagePlayer( "Error - Location name already is in use!.", player );
        else
        {
			SQLite_Exec( db, "INSERT INTO Gotoloc (Name, x, y, z, Creator) values ( '" + text + "', '" + player.Pos.x + "', '" + player.Pos.y + "', '" + player.Pos.z + "', '" + player.Name + "' )" );
            MessagePlayer( "Location: " + text + " Saved!", player );                                       
        }
                     
	}
	
	else if (cmd == "gotoloc")
    {          
		if ( !text ) MessagePlayer( "Error - Usage: / gotoloc <locname>", player );
		else if ( !player.IsSpawned ) MessagePlayer( "Error - You haven't spawned.", player );
		else
        {
			local q = SQLite_Query( db, "SELECT x, y, z, Creator FROM Gotoloc WHERE Name = '"+ text +"'" );
			if ( q )
			{
				local x = SQLite_ColumnData( q, 0 ), y = SQLite_ColumnData( q, 1 ), z =SQLite_ColumnData( q, 2 ), Creator = SQLite_ColumnData( q, 3 );
				player.Pos = Vector( x, y, z );
			}
			else MessagePlayer("Error - Invalid location", player);
			SQLite_Release( q );
        }
    }
	 
	return 1;
}

function AutoSave()
{
	if( autoSave && CurrentMap != "" && objCreated > 1 )
	{
		SQLite_Exec( Maps, "DELETE FROM '" + CurrentMap.tolower() + "'" );
		for( local i = 0; i < objCreated; i++ )
		{
			local obj = FindObject(i);
			if( obj )
			{
				SQLite_Exec( Maps, "INSERT INTO '" + CurrentMap.tolower() + "' VALUES ( '" + obj.Model + "', '" + obj.Pos.x + "', '" + obj.Pos.y + "', '" + obj.Pos.z + "', '" + obj.RotationEuler.x + "', '" + obj.RotationEuler.y + "', '" + obj.RotationEuler.z + "');" );
			}
		}
		Message( "[#00CC00][AUTO-SAVE]: [#ffffff]Map has been saved!" );
	}
	else return;
}

function ClientRender( idPlayer )
{
	local player = FindPlayer( idPlayer );
	if( player )
	{
		txdCurrentMap.Delete();
		txdTotalObjects.Delete();
		
		txdTotalObjects = CreateTextdraw( "Total Objects Created: " + objCreated, 1, 720, 0xFF00FFFF );
		txdTotalObjects.SetRelativeForAll( true );
		txdTotalObjects.ShowForPlayer( player );
		
		if( CurrentMap != "" ) txdCurrentMap = CreateTextdraw( "Current Map: " + CurrentMap, 1, 735, 0xFFFFFF00 );
		else txdCurrentMap = CreateTextdraw( "Current Map: None" 1, 735, 0xFFFFFF00 );
		txdCurrentMap.SetRelativeForAll( true );
		txdCurrentMap.ShowForPlayer( player );
		
		if( pData[ player.ID ].txdObject != null ) 
		{ 
			pData[ player.ID ].txdObject.Delete(); 
			pData[ player.ID ].txdObject = null; 
		}
	
		if( Object[ player.ID ] == null )
		{
			pData[ player.ID ].txdObject = CreateTextdraw( "Object: None", 1, 750, 0xFF00FF00 );
			pData[ player.ID ].txdObject.SetRelativeForAll( true );
			pData[ player.ID ].txdObject.ShowForPlayer( player );
		}
		else
		{
			local object; 
			if( typeof Object[ player.ID ] == "instance" ) object = Object[ player.ID ];
			else object = FindObject( Object[ player.ID ] );
			
			local objectModel = object.Model, objectID = object.ID;
			
			pData[ player.ID ].txdObject = CreateTextdraw( "Object: Model [ " + objectModel + " ] ID ( " + objectID + " )", 1, 750, 0xFF00FF00 );
			pData[ player.ID ].txdObject.SetRelativeForAll( true );
			pData[ player.ID ].txdObject.ShowForPlayer( player );
		}	
	}	
}

function FindLoc( text ) 
{
        local Location = SQLite_Query( db, "SELECT Name FROM Gotoloc WHERE Name='" + text + "'" );
        if( Location ) { SQLite_Release( Location ); return true }
        else { SQLite_Release( Location ); return false }
}

function onObjectShot( object, player, weapon )
{
	local obj = FindObject( object.ID );
	local plr = FindPlayer( player.ID );
	Object[ plr.ID ] = obj;
	pData[ player.ID ].Editing = true;
	MessagePlayer( "[#00CC00][SELECT]: [#ffffff]Selected object " + object.ID, player );
	return 1;
}

function onObjectBump( object, player )
{
	local obj = FindObject( object.ID );
	local plr = FindPlayer( player.ID );
	obj.Delete();
	pData[ player.ID ].Editing = false;
	objCreated--;
	Object[ player.ID ] = null;
	MessagePlayer( "[#00CC00][SELECT]: [#ffffff]Object deleted!", player );
}

function EnableBumpTracks()
{
	for( local i = 0; i < objCreated; i++ )
	{
		local o = FindObject( i );
		if( o ) o.TrackingBumps = true;
	}
}

function DisableBumpTracks()
{
	for( local i = 0; i < objCreated; i++ )
	{
		local o = FindObject( i );
		if( o ) o.TrackingBumps = false;
	}
}

function RemoveObjects()
{
	for( local i = 0; i < objCreated; i++ )
	{
		local obj = FindObject( i );
		if( obj ) { obj.Delete(); objCreated--; }
	}
	while( objCreated != 0 );
	objCreated = 0;
	Message( "[#ffffff]All objects of the map " + CurrentMap + " were removed!" );
}

function onKeyUp( player, key )
{
	if( pData[ player.ID ].Motion != null )
	{
		pData[ player.ID ].Motion.Delete();
		pData[ player.ID ].Motion = null;
	}
	
	if( key == ctrl ) { pData[ player.ID ].CTRL = false; }
	onCamKeyUp( player, key );
}

function onKeyDown( plr, key )
{
	local player;
	if( typeof plr != "instance" ) player = FindPlayer( plr );
	else player = plr;
	
	if( !pData[ player.ID ].Editing && pData[ player.ID ].lastPos == null ) return onCamKeyDown( plr, key ); /*MessagePlayer( "[#00CC00][MAP]: [#ffffff]You are not controlling any object!", player );*/
	
	if( key == ArrowRight && pData[ player.ID ].lastPos != null )
	{
		pData[ player.ID ].viewObject.Delete();
		pData[ player.ID ].viewObject = null;
		pData[ player.ID ].viewModel++;
		pData[ player.ID ].viewObject = CreateObject( pData[ player.ID ].viewModel, player.World, -1372.88, 1282.51, 83.8704, 255 );
		//MessagePlayer( "[#00CC00][OBJECT]: [#FFFFFF]Viewing object: " + pData[ player.ID ].viewModel, player );
	}
	else if( key == ArrowLeft && pData[ player.ID ].lastPos != null )
	{
		pData[ player.ID ].viewObject.Delete();
		pData[ player.ID ].viewObject = null;
		pData[ player.ID ].viewModel--;
		pData[ player.ID ].viewObject = CreateObject( pData[ player.ID ].viewModel, player.World, -1372.88, 1282.51, 83.8704, 255 );
		//MessagePlayer( "[#00CC00][OBJECT]: [#FFFFFF]Viewing object: " + pData[ player.ID ].viewModel, player );
	}
	else if( key == ctrl && pData[ player.ID ].lastPos != null )
	{
		pData[ player.ID ].motionTimer.Delete();
		player.Pos = pData[ player.ID ].lastPos;
		pData[ player.ID ].lastPos = null;
		player.RestoreCamera();
		player.World = 0;
		pData[ player.ID ].viewObject.Delete();
		pData[ player.ID ].viewObject = null;
		player.IsFrozen = false;
		Object[ player.ID ] = CreateObject( pData[ player.ID ].viewModel, 0, player.Pos.x, player.Pos.y, player.Pos.z, 255 );
		Object[ player.ID ].TrackingShots = true;
		objCreated++;
		player.Pos = Vector( player.Pos.x, player.Pos.y, player.Pos.z + 3 );
		pData[ player.ID ].Editing = true;
		Message( "[#00CC00][MAP]: [#ffffff]Created object " + pData[ player.ID ].viewModel );
	}
	
	if( pData[ player.ID ].Motion == null && key != one && key != R && key != two && key != backspace && key != del ) pData[ player.ID ].Motion = NewTimer( "onKeyDown", Motion_Speed, 0, player.ID, key );
	
	if( key == ArrowLeft && !pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, 0.05 ), 100 );
			break;
			
			case "fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, 0.30 ), 100 );
			break;
			
			case "very fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, 1.30 ), 100 );
			break;
			
			case "very slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, 0.01 ), 100 );
			break;
			
			case "slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, 0.03 ), 100 );
			break;
		}
		return;
	}
	
	else if( key == ArrowRight && !pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, -0.05 ), 100 );
			break;
			
			case "fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, -0.30 ), 100 );
			break;
			
			case "very fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, -1.30 ), 100 );
			break;
			
			case "very slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, -0.01 ), 100 );
			break;
			
			case "slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0, -0.03 ), 100 );
			break;
		}
		return;
	}
	
	else if( key == PageUp && !pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].RotateByEuler( Vector( 0.05, 0, 0 ), 100 );
			break;
			
			case "fast":
			Object[ player.ID ].RotateByEuler( Vector( 0.30, 0, 0 ), 100 );
			break;
			
			case "very fast":
			Object[ player.ID ].RotateByEuler( Vector( 1.30, 0, 0 ), 100 );
			break;
			
			case "very slow":
			Object[ player.ID ].RotateByEuler( Vector( 0.01, 0, 0 ), 100 );
			break;
			
			case "slow":
			Object[ player.ID ].RotateByEuler( Vector( 0.03, 0, 0 ), 100 );
			break;
		}
		return;
	}
	
	else if( key == PageDown && !pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].RotateByEuler( Vector( -0.05, 0, 0 ), 100 );
			break;
			
			case "fast":
			Object[ player.ID ].RotateByEuler( Vector( -0.30, 0, 0 ), 100 );
			break;
			
			case "very fast":
			Object[ player.ID ].RotateByEuler( Vector( -1.30, 0, 0 ), 100 );
			break;
			
			case "very slow":
			Object[ player.ID ].RotateByEuler( Vector( -0.01, 0, 0 ), 100 );
			break;
			
			case "slow":
			Object[ player.ID ].RotateByEuler( Vector( -0.03, 0, 0 ), 100 );
			break;
		}
		return;
	}
	
	else if( key == ArrowUp && !pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0.05, 0 ), 100 );
			break;
			
			case "fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0.30, 0 ), 100 );
			break;
			
			case "very fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, 1.30, 0 ), 100 );
			break;
			
			case "very slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0.01, 0 ), 100 );
			break;
			
			case "slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, 0.03, 0 ), 100 );
			break;
		}
		return;
	}
	
	else if( key == ArrowDown && !pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].RotateByEuler( Vector( 0, -0.05, 0 ), 100 );
			break;
			
			case "fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, -0.30, 0 ), 100 );
			break;
			
			case "very fast":
			Object[ player.ID ].RotateByEuler( Vector( 0, -1.30, 0 ), 100 );
			break;
			
			case "very slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, -0.01, 0 ), 100 );
			break;
			
			case "slow":
			Object[ player.ID ].RotateByEuler( Vector( 0, -0.03, 0 ), 100 );
			break;
		}
		return;
	}
	
	else if( key == PageUp && pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		local x = Object[ player.ID ].Pos.x;
		local y = Object[ player.ID ].Pos.y;
		local z = Object[ player.ID ].Pos.z;
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].Pos = Vector( x, y, z + 0.05 );
			break;
			
			case "fast":
			Object[ player.ID ].Pos = Vector( x, y, z + 0.30 );
			break;
			
			case "very fast":
			Object[ player.ID ].Pos = Vector( x, y, z + 1.30 );
			break;
			
			case "very slow":
			Object[ player.ID ].Pos = Vector( x, y, z + 0.01 );
			break;
			
			case "slow":
			Object[ player.ID ].Pos = Vector( x, y, z + 0.03 );
			break;
		}
		return;
	}
	
	
	else if( key == PageDown && pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		local x = Object[ player.ID ].Pos.x;
		local y = Object[ player.ID ].Pos.y;
		local z = Object[ player.ID ].Pos.z;
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].Pos = Vector( x, y, z - 0.05 );
			break;
			
			case "fast":
			Object[ player.ID ].Pos = Vector( x, y, z - 0.30 );
			break;
			
			case "very fast":
			Object[ player.ID ].Pos = Vector( x, y, z - 1.30 );
			break;
			
			case "very slow":
			Object[ player.ID ].Pos = Vector( x, y, z - 0.01 );
			break;
			
			case "slow":
			Object[ player.ID ].Pos = Vector( x, y, z - 0.03 );
			break;
		}
		return;
	}
	
	else if( key == ArrowLeft && pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		local x = Object[ player.ID ].Pos.x;
		local y = Object[ player.ID ].Pos.y;
		local z = Object[ player.ID ].Pos.z;
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].Pos = Vector( x - 0.05, y, z );
			break;
			
			case "fast":
			Object[ player.ID ].Pos = Vector( x - 0.30, y, z );
			break;
			
			case "very fast":
			Object[ player.ID ].Pos = Vector( x - 1.30, y, z);
			break;
			
			case "very slow":
			Object[ player.ID ].Pos = Vector( x - 0.01, y, z );
			break;
			
			case "slow":
			Object[ player.ID ].Pos = Vector( x - 0.03, y, z );
			break;
		}
		return;
	}
	
	else if( key == ArrowRight && pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		local x = Object[ player.ID ].Pos.x;
		local y = Object[ player.ID ].Pos.y;
		local z = Object[ player.ID ].Pos.z;
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].Pos = Vector( x + 0.05, y, z );
			break;
			
			case "fast":
			Object[ player.ID ].Pos = Vector( x + 0.30, y, z );
			break;
			
			case "very fast":
			Object[ player.ID ].Pos = Vector( x + 1.30, y, z);
			break;
			
			case "very slow":
			Object[ player.ID ].Pos = Vector( x + 0.01, y, z );
			break;
			
			case "slow":
			Object[ player.ID ].Pos = Vector( x + 0.03, y, z );
			break;
		}
		return;
	}
	
	else if( key == ArrowUp && pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		local x = Object[ player.ID ].Pos.x;
		local y = Object[ player.ID ].Pos.y;
		local z = Object[ player.ID ].Pos.z;
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].Pos = Vector( x, y + 0.05, z );
			break;
			
			case "fast":
			Object[ player.ID ].Pos = Vector( x, y + 0.30, z );
			break;
			
			case "very fast":
			Object[ player.ID ].Pos = Vector( x, y + 1.30, z);
			break;
			
			case "very slow":
			Object[ player.ID ].Pos = Vector( x, y + 0.01, z );
			break;
			
			case "slow":
			Object[ player.ID ].Pos = Vector( x, y + 0.03, z );
			break;
		}
		return;
	}
	
	else if( key == ArrowDown && pData[ player.ID ].PositionMode && Object[ player.ID ] != null )
	{
		local x = Object[ player.ID ].Pos.x;
		local y = Object[ player.ID ].Pos.y;
		local z = Object[ player.ID ].Pos.z;
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			Object[ player.ID ].Pos = Vector( x, y - 0.05, z );
			break;
			
			case "fast":
			Object[ player.ID ].Pos = Vector( x, y - 0.30, z );
			break;
			
			case "very fast":
			Object[ player.ID ].Pos = Vector( x, y - 1.30, z);
			break;
			
			case "very slow":
			Object[ player.ID ].Pos = Vector( x, y - 0.01, z );
			break;
			
			case "slow":
			Object[ player.ID ].Pos = Vector( x, y - 0.03, z );
			break;
		}
		return;
	}
	
	else if( key == one && Object[ player.ID ] != null )
	{
		if( !pData[ player.ID ].PositionMode )
		{
			pData[ player.ID ].PositionMode = true;
			MessagePlayer( "[#ffffff]Map positioning mode is [#00ff00]ON[#ffffff]! Rotation mode [#ff0000]OFF[#ffffff]!", player );
		}
		
		else if( pData[ player.ID ].PositionMode )
		{
			pData[ player.ID ].PositionMode = false;
			MessagePlayer( "[#ffffff]Map positioning mode is [#ff0000]OFF[#ffffff]! Rotation mode [#00ff00]ON[#ffffff]!", player );
		}
	}
	
	else if( key == two && Object[ player.ID ] != null )
	{
		switch( pData[ player.ID ].Speed )
		{
			case "normal":
			pData[ player.ID ].Speed = "fast";
			MessagePlayer( "[#ffffff]Speed movement level is now [#00ff00]Fast[#ffffff]!", player );
			break;
			
			case "fast":
			pData[ player.ID ].Speed = "very fast";
			MessagePlayer( "[#ffffff]Speed movement level is now [#00ff00]Very Fast[#ffffff]!", player );
			break;
			
			case "very fast":
			pData[ player.ID ].Speed = "very slow";
			MessagePlayer( "[#ffffff]Speed movement level is now [#00ff00]Very Slow[#ffffff]!", player );
			break;
			
			case "very slow":
			pData[ player.ID ].Speed = "slow";
			MessagePlayer( "[#ffffff]Speed movement level is now [#00ff00]Slow[#ffffff]!", player );
			break;
			
			case "slow":
			pData[ player.ID ].Speed = "normal";
			MessagePlayer( "[#ffffff]Speed movement level is now [#00ff00]Normal[#ffffff]!", player );
			break;
		}
	}
	
	else if( key == ctrl ) { pData[ player.ID ].CTRL = true; }
	
	else if( key == c && pData[ player.ID ].CTRL && Object[ player.ID ] != null )
	{
		local old_obj = Object[ player.ID ].Pos;
		local old_objid = Object[ player.ID ].Model;
		local old_objrot = Object[ player.ID ].RotationEuler;
		
		Object[ player.ID ] = CreateObject( old_objid, 0, old_obj, 255 );
		Object[ player.ID ].RotateToEuler( old_objrot, 0 );
		Object[ player.ID ].TrackingShots = true;
		objCreated++;
		pData[ player.ID ].Editing = true;
		Message( "[#00CC00][CLONE]: [#ffffff]Cloned object " + old_objid + " at its position" );
	}
	
	else if( key == del && Object[ player.ID ] != null )
	{
		local object;
		if( typeof Object[ player.ID ] != "instance" ) object = FindObject( Object[ player.ID ] );
		else object = Object[ player.ID ];
		if( !object ) return MessagePlayer( "[#00CC00][ERROR]: [#FFFFFF]Could not delete object", player );
		object.Delete();
		Object[ player.ID ] = null;
		pData[ player.ID ].Editing = false;
		objCreated--;
		MessagePlayer( "[#00CC00][SELECT]: [#ffffff]Object deleted!", player );
	}
	
	else if( key == R && Object[ player.ID ] != null )
	{
		Object[ player.ID ].RotateToEuler( Vector( 0, 0, 0 ), 0 );
		MessagePlayer( "[#ffffff]Rotation angles have been Reset!", player );
		return;
	}
	
	else if( key == backspace && pData[ player.ID ].Editing )
	{
		pData[ player.ID ].Editing = false;
		Object[ player.ID ] = null;
		MessagePlayer( "[#00CC00][EDIT]: [#ffffff]You have finished editing the object.", player );
	}
}

function onPlayerHealthChange( player, lastHP, newHP )
{
	player.Health = 100;
}

function onObjectRender( idPlayer )
{
	local player = FindPlayer( idPlayer );
	if( player && pData[ player.ID ].viewObject != null )
	{
		local object = pData[ player.ID ].viewObject;
		if( object ) 
		{
			if( object.World != player.World )
			{
				MessagePlayer( "Not same world, fixing.", player );
				object.World = player.World;
			}
			object.RotateByEuler( Vector( 0, 0, -0.1 ), 100 );
		}
	}
	else return null;
}
// ================================== END ======================================